/* 
 * config file parser functions
 *
 * (C) 2000-2003 by Harald Welte <laforge@gnumonks.org>
 * (C) 2004 by Michal Kwiatkowski <ruby@joker.linuxstuff.pl>
 */

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 
 *  as published by the Free Software Foundation
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <conffile/conffile.h>


#ifdef DEBUG
#define DEBUGP(format, args...) fprintf(stderr, "%s:%i:" format, __FILE__, __LINE__, ## args)
#else
#define DEBUGP(format, args...)
#endif


#define BLOCK_NAME_LEN 8
typedef struct block {
	struct block *	next;
	char		name[BLOCK_NAME_LEN];
	long		begin,
			end;
	short int	line,
			flag; /* set to 1 if block was accessed by config_block() */
} block_t;


static FILE *fd;
static block_t *blocks;
static block_t *current_block;


/* 
 * get_word()  - parse a line into words
 * 
 * Arguments:     line   line to parse
 *                delim  possible word delimiters
 *                buf    pointer to buffer where word is returned
 * Return value:         pointer to first char after word
 * 
 * This function can deal with "" quotes.
 *    Written by Harald Welte, modified by Michal Kwiatkowski.
 */
static char *get_word(char *line, char *delim, char *buf)
{
	char *p, *start = NULL, *stop = NULL;
	int inquote = 0;

	for (p = line; *p; p++) {
		if (*p == '"') {
			start  = p + 1;
			inquote = 1;
			break;
		}
		if (!strchr(delim, *p)) {
			start = p;
			break;
		}
	}
	if (!start)
		return NULL;

	/* determine pointer to one char after word */
	for (p = start; *p; p++) {
		if (inquote) {
			if (*p == '"') {
				stop = p;
				break;
			}
		} else {
			if (strchr(delim, *p)) {
				stop = p;
				break;
			}
		}
	}
	if (!stop)
		return NULL;

	if (buf) {
		strncpy(buf, start, (size_t) (stop-start));
		*(buf + (stop-start)) = '\0';
	}

	/* skip quote character */
	if (inquote)
		/* yes, we can return stop + 1. If " was the last 
		 * character in string, it now points to NULL-term */
		return (stop + 1);

	return stop;
}


/*
 * read_line()  - reads one line from file and snips comments
 * 
 * Arguments:     line  buffer to store line
 * Return value:        number of lines read
 */
static int __lines_read;
static int read_line(char *line, int len)
{
	char *p;

	if (fgets(line, len, fd) == NULL) {
		return 0;
	}
	__lines_read++;

	/* comment should act as new line and string delimiter */
	if ((p = strchr(line, '#')) != NULL) {
		*p = '\n';
		*(p+1) = '\0';
	}

	/* look for line continuation: escape and a newline */
	if ((p = strstr(line, "\\\n")) != NULL) {
		/* recursion - the easiest way */
		*p = '\0';
		len -= strlen(line);
		if (len <= 0)
			return 0;
		if (read_line(p, len) == 0)
			return 0;
	}

	return __lines_read;
}


/*
 * find_key()  - looks for config entry with given key name in list
 * 
 * Arguments:     name  name of option we're looking for
 *                list  list of config entries to check
 * Return value:        pointer to config entry or NULL if not found
 */
static config_entry_t *find_key(char *name, config_entry_t *list)
{
	config_entry_t *ce;

	for (ce = list; ce; ce = ce->next) {
		if (!strncmp(name, ce->key, CONFIG_KEY_LEN))
			return ce;
	}

	return NULL;
}


/*
 * mtoi()  - changes memory string into plain integer
 *
 * Arguments:     str   string to parse
 * Return value:        integer value of given string
 */
static int mtoi(char *str)
{
	char number[32];
	char *last;
	int factor = 1;

	strncpy(number, str, 31);
	last = &(number[strlen(number) - 1]);
	switch (*last) {
		case 'M':
			factor *= 1024;
		case 'K':
			factor *= 1024;
			*last = '\0';
		default:
			break;
	}

	return (atoi(number) * factor);
}


/* 
 * config_open()  - opens a file and divide it into blocks
 * 
 * Arguments:     path  points to config file
 * Return value:        number of blocks found
 *
 * Block starts with string and curly bracket '{', while ends with '}'.
 * No nesting is allowed. '#' is treated as a comment.
 */
int config_open(char *path)
{
	int found = 0, line_ctr = 0, inside_block = 0;

	if ((fd = fopen(path, "r")) == NULL) {
		perror("Couldn't open config file");
		return 0;
	}

	/* search for blocks */
	do {
		int ret;
		char line[CONFIG_LINE_LEN];
		char id[CONFIG_VALUE_LEN], bracket[CONFIG_VALUE_LEN];
		char *p;
		block_t *bl;

		/* block can end at the same line it started,
		 * so let's check for this first */
		if (inside_block) {
			char *s;
			if ((s = strchr(line, '}')) != NULL) {
				if (get_word(s+1, " \t\n", NULL)) {
					fprintf(stderr, "Config error at line %d: "
						"trailing characters after end of block.\n",
						line_ctr);
					return 0;
				}
				bl->end = ftell(fd) - strlen(line) + (s - line);
				found++;
				inside_block=0;
			}
			if (strchr(line, '{')) {
				fprintf(stderr, "Config error at line %d: "
						"nesting of blocks forbidden.\n",
						line_ctr);
				return 0;
			}
		}

		__lines_read = 0;
		if ((ret = read_line(line, CONFIG_LINE_LEN-1)) == 0)
			break;
		line_ctr += ret;

		/* if we're inside block, just look for its end,
		 * validity will be checked during config_read() */
		if (inside_block)
			continue;

		/* read two words, second should be curly bracket */
		p  = line;
		if ((p = get_word(p, " \t\n", id)) == NULL)
			continue;
		p = get_word(p, " \t\n", bracket);
		if (p == NULL || bracket[0] != '{') {
			fprintf(stderr, "Config error at line %d: "
					"\"%s\" is not a block.\n",
					line_ctr, id);
			return 0;
		}

		for (bl = blocks; bl; bl = bl->next) {
			if (!strcmp(bl->name, id)) {
				fprintf(stderr, "Config error at line %d: "
						"redefinition of block \"%s\".\n",
						line_ctr, id);
				return 0;
			}
		}

		/* create new block_t structure and fill it */
		if ((bl = (block_t *)malloc(sizeof(block_t))) == NULL) {
			fprintf(stderr, "config_open(): out of memory.\n");
			return 0;
		}
		if (!blocks)
			blocks = bl;
		else {
			block_t *tmp = blocks;
			while (tmp->next)
				tmp = tmp->next;
			tmp->next = bl;
		}
		bl->next = NULL;
		strncpy(bl->name, id, BLOCK_NAME_LEN);
		bl->begin = ftell(fd) - (strlen(line) - (strchr(line, '{') - line + 1));
		bl->line = line_ctr;
		bl->flag = 0;
		inside_block = 1;
		/* clear start of block, so it won't drop into
		 * condition at line 241 */
		*(strchr(line, '{')) = ' ';

	} while (!feof(fd));

	if (inside_block) {
		fprintf(stderr, "Config error at line %d: "
				"unexpected end of file.\n", line_ctr);
		return 0;
	}

	return found;
}


/*
 * config_block()  - sets a current block and sets file position
 *                   to the beginning of it
 *
 * Arguments:     name  name of requested block
 * Return value:        -1 if no block with given name was found
 */
int config_block(char *name)
{
	block_t *p;

	for (p = blocks; p; p = p->next) {
		if (!strcmp(p->name, name)) {
			current_block = p;
			fseek(fd, p->begin, SEEK_SET);
			p->flag = 1;
			return 0;
		}
	}

	return -1;
}


/*
 * config_read()  - read options from current block
 *
 * Arguments:     list  list of config entries to fill
 * Return value:        -1 on error, 0 if whole block was read,
 *                      1 if there are still some options to parse
 *
 * If list->options have CONFIG_OPT_MULTI set, only this options
 * is looked for, others are ignored.
 *
 * Function normally returns after parsing all options or after reading
 * entry with CONFIG_OPT_MULTI set. In the second case function must
 * be invoked many times to read all multi options.
 */
int config_read(config_entry_t *list)
{
	config_entry_t *ce;
	int line_ctr;
	char line[CONFIG_LINE_LEN];
	char wordbuf[CONFIG_VALUE_LEN];
	char *wordend, *p;

	if (!current_block) {
		DEBUGP("config_read(): Call config_block() first!\n");
		return -1;
	}

	line_ctr = current_block->line - 1;

	while (ftell(fd) < current_block->end) {
		int ret;

		__lines_read = 0;
		if ((ret = read_line(line, CONFIG_LINE_LEN-1)) == 0)
			break;
		line_ctr += ret;

		/* terminate at end of block */
		if ((p = strchr(line, '}')) != NULL) {
			*p = '\n';
			*(p+1) = '\0';
		}

		if ((wordend = get_word(line, " \t\n", wordbuf)) == NULL)
			continue;

		if ((ce = find_key(wordbuf, list)) == NULL) {
			if (list->options != CONFIG_OPT_MULTI) {
				fprintf(stderr, "Config error at line %d: "
						"bad option \"%s\".\n",
						line_ctr, wordbuf);
				return -1;
			}
			else {
				continue;
			}
		}

		if (ce->options == CONFIG_OPT_IGNORE)
			continue;

		if (ce->hit) {
			fprintf(stderr, "Config error at line %d: "
					"redefinition of option \"%s\".\n",
					line_ctr, wordbuf);
			return -1;
		}

		if (ce->type != CONFIG_TYPE_BOOLEAN) {
			if ((wordend = get_word(wordend, " \t\n", wordbuf)) == NULL) {
				fprintf(stderr, "Config error at line %d: "
					"no value for option \"%s\".\n",
					line_ctr, wordbuf);
				return -1;
			}
		}

		if (get_word(wordend, " \t\n", wordbuf) != NULL) {
			fprintf(stderr, "Config error at line %d: "
					"trailing characters after option.\n",
					line_ctr);
			return -1;
		}

		switch (ce->type) {
			case CONFIG_TYPE_BOOLEAN:
				ce->u.value = 1;
				break;
			case CONFIG_TYPE_INT:
				ce->u.value = strtol(wordbuf, NULL, 10);
				break;
			case CONFIG_TYPE_HEX:
				ce->u.value = strtol(wordbuf, NULL, 16);
				break;
			case CONFIG_TYPE_STRING:
				strncpy(ce->u.string, wordbuf, CONFIG_VALUE_LEN);
				break;
			case CONFIG_TYPE_MEM:
				ce->u.value = mtoi(wordbuf);
				break;
			default:
				DEBUGP("config_read(): bad config type %i.\n", ce->type);
				break;
		}

		if (ce->options == CONFIG_OPT_MULTI)
			return 1;
		else
			ce->hit = 1;
	}

	for (ce = list; ce; ce = ce->next) {
		if ((ce->options == CONFIG_OPT_MANDATORY) && !ce->hit) {
			fprintf(stderr, "Config error in block \"%s\": "
					"mandatory option \"%s\" not set.\n",
					current_block->name, ce->key);
			return -1;
		}
	}

	return 0;
}


/*
 * config_close()  - closes config file and frees all block descriptors
 *
 * Arguments:     none
 * Return value:        -1 on error
 */
int config_close(void)
{
	block_t *bl;

	if (fclose(fd) == EOF) {
		DEBUGP("config_close(): Couldn't close config file.\n");
		return -1;
	}

	for (bl = blocks; bl; bl = bl->next) {
		if (!bl->flag) {
			fprintf(stderr, "Config error at line %d: "
					"bad block id \"%s\".\n",
					bl->line, bl->name);
			return -1;
		}
	}

	while (blocks) {
		bl = blocks->next;
		free(blocks);
		blocks = bl;
	}
	current_block = NULL;

	return 0;
}

