/* specter_EXEC.c
 *
 * specter output target for execution of applications
 *
 * This target executes application on packet receive, expanding few
 *   conversion tags into values appropriate for current packet.
 *
 * (C) 2004 by Michal Kwiatkowski <ruby@joker.linuxstuff.pl>
 */

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 
 *  as published by the Free Software Foundation
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <specter/specter.h>
#include <conffile/conffile.h>
#include "lret.h"


/* maximym length of whole command line */
#ifndef EXEC_MAX_LINE
#define EXEC_MAX_LINE 256
#endif

/* maximum number of command line arguments including application name */
#ifndef EXEC_MAX_ARGS
#define EXEC_MAX_ARGS 32
#endif

/* space for single argument during substitution */
#ifndef EXEC_ARG_SPACE
#define EXEC_ARG_SPACE 32
#endif


static config_entry_t my_config[] = {
	{ .key = "execstr", .type = CONFIG_TYPE_STRING, .options = CONFIG_OPT_MANDATORY },
	{ .key = "execwait", .type = CONFIG_TYPE_BOOLEAN, .options = CONFIG_OPT_NONE,
		.u.value = 0 },
	{ .key = "execforce", .type = CONFIG_TYPE_BOOLEAN, .options = CONFIG_OPT_NONE,
		.u.value = 0 },
};


struct my_data {
	char execline[EXEC_MAX_LINE];
	char *argv[EXEC_MAX_ARGS+1];
};


/* here goes conversion stuff */
#define LOCAL_RET_NUM 10
struct {
	char name[SPECTER_IRET_NAME_LEN];
	specter_iret_t *p;
	char c;
} local_ret[LOCAL_RET_NUM] = {
/*  0 */{ "oob.in", NULL, 'I' },
	{ "oob.out", NULL, 'O' },
	{ "ip.saddr", NULL, 'S' },
	{ "ip.daddr", NULL, 'D' },
	{ "ip.protocol", NULL, 'P' },
/*  5 */{ "tcp.sport", NULL, 's' },
	{ "tcp.dport", NULL, 'd' },
	{ "udp.sport", NULL, 's' },
	{ "udp.dport", NULL, 'd' },
	{ "icmp.type", NULL, 'i' },
};


static specter_iret_t *find_tag(const char c)
{
	int ctr;
	specter_iret_t *iret = NULL;

	for (ctr = 0; ctr < LOCAL_RET_NUM; ctr++)
		if (local_ret[ctr].c == c) {
			iret = local_ret[ctr].p;
			if (IS_VALID(iret))
				break;
		}

	return iret;
}

static int fill_tag(const char c)
{
	int ctr;

	for (ctr = 0; ctr < LOCAL_RET_NUM; ctr++)
		if (local_ret[ctr].c == c) {
			if ((local_ret[ctr].p = find_iret(local_ret[ctr].name)) != NULL)
				return 0;
		}

	return -1;
}


static void check_status(char *comm, pid_t pid, int status)
{
	if (WIFEXITED(status)) {
		if (WEXITSTATUS(status)) {
			specter_log(SPECTER_NOTICE, "%s[%i] returned %i.\n",
					comm, pid, WEXITSTATUS(status));
		}
		else {
			specter_log(SPECTER_DEBUG, "%s[%i] returned 0.\n",
					comm, pid);
		}
	}
	else if (WIFSIGNALED(status)) {
		specter_log(SPECTER_NOTICE, "%s[%i] terminated by signal #%i.\n",
					comm, pid, WTERMSIG(status));
	}
}


static void exec_fini(config_entry_t *ce, void *data)
{
	int ret, status;
	struct my_data *md = data;

	while ((ret = waitpid(-1, &status, WNOHANG)) > 0) {
		check_status(md->argv[0], ret, status);
	}

	free(data);
}


static void *exec_init(config_entry_t *ce)
{
	struct my_data *data;
	char *tmp;
	int i;
	struct stat st;

	/* check if there're only allowed tags in execline */
	tmp = GET_CE(ce,0)->u.string;
	while ((tmp = strchr(tmp, '%')) != NULL) {
		tmp++;
		if (*tmp == '%') {
			tmp++;
			continue;
		}
		if (fill_tag(*tmp) == -1) {
			specter_log(SPECTER_FATAL, "Bad exec tag: \"%%%c\".\n", *tmp);
			return NULL;
		}
		if (!strchr(" \t\n\0", *(tmp+1))) {
			specter_log(SPECTER_FATAL, "Only one-character tags allowed.\n");
			return NULL;
		}
	}

	if ((data = malloc(sizeof(struct my_data))) == NULL) {
		specter_log(SPECTER_FATAL, "Couldn't allocate data: %s.\n",
				strerror(errno));
		return NULL;
	}

	memset(data->argv, 0x0, (EXEC_MAX_ARGS+1)*sizeof(char *));

	/* create argv array for execve() */
	tmp = strncpy(data->execline, GET_CE(ce,0)->u.string, EXEC_MAX_LINE);
	for (i = 0; *tmp && i < EXEC_MAX_ARGS; i++) {
		while (*tmp == ' ')
			*tmp++ = '\0';
		data->argv[i] = tmp;
		while (*tmp != ' ' && *tmp != '\0')
			tmp++;
	}

	/* check if file exists and is an executable */
	if (stat(data->argv[0], &st) == -1) {
		specter_log(SPECTER_FATAL, "Couldn't stat(%s): %s.\n",
				data->argv[0], strerror(errno));
		exit(EXIT_FAILURE);
	}
	if (!(st.st_mode & (S_IXOTH|S_IXGRP|S_IXUSR)) || !(st.st_mode & S_IFREG)) {
		specter_log(SPECTER_FATAL, "\"%s\" isn't an executable.\n",
				data->argv[0]);
		exit(EXIT_FAILURE);
	}

	return data;
}


static int output_exec(config_entry_t *ce, void *data)
{
	static char buff[EXEC_MAX_ARGS][EXEC_ARG_SPACE];
	static char *argv[EXEC_MAX_ARGS+1];
	int i, ctr, status;
	struct my_data *md = data;

	/* first check if any previously executed children returned */
	while ((i = waitpid(-1, &status, WNOHANG)) > 0) {
		check_status(md->argv[0], i, status);
	}

	/* let's work on a copy */
	memcpy(argv, md->argv, (EXEC_MAX_ARGS+1)*sizeof(char *));

	/* resolve tags */
	for (ctr = 1, i = 0; argv[ctr]; ctr++) {
		char *p;
		specter_iret_t *iret;

		if ((p = strchr(argv[ctr], '%')) == NULL)
			continue;
		p++;
		if (*p == '%')
			continue;

		if ((iret = find_tag(*p)) == NULL) {
			specter_log(SPECTER_DEBUG, "Unknown tag: %c.\n", *p);
			return -1;
		}

		if (IS_VALID(iret)) {
			if (i == EXEC_MAX_ARGS) {
				specter_log(SPECTER_DEBUG, "Not enough space for arguments.\n");
				return -1;
			}

			switch (iret->type) {
				case SPECTER_IRET_INT8:
					sprintf(buff[i], "%d", iret->value.i8);
					break;
				case SPECTER_IRET_INT16:
					sprintf(buff[i], "%d", iret->value.i16);
					break;
				case SPECTER_IRET_INT32:
					sprintf(buff[i], "%d", iret->value.i32);
					break;
				case SPECTER_IRET_INT64:
					sprintf(buff[i], "%lld", iret->value.i64);
					break;
				case SPECTER_IRET_UINT8:
					sprintf(buff[i], "%u", iret->value.ui8);
					break;
				case SPECTER_IRET_UINT16:
					sprintf(buff[i], "%u", iret->value.ui16);
					break;
				case SPECTER_IRET_UINT32:
					sprintf(buff[i], "%u", iret->value.ui32);
					break;
				case SPECTER_IRET_UINT64:
					sprintf(buff[i], "%llu", iret->value.ui64);
					break;
				case SPECTER_IRET_BOOL:
					sprintf(buff[i], "%d", iret->value.b);
					break;
				case SPECTER_IRET_IPADDR:
					{
					struct in_addr ia;
					char *str;

					ia.s_addr = ntohl(iret->value.ui32);
					str = inet_ntoa(ia);

					if (strlen(str) > EXEC_ARG_SPACE - 1) {
						specter_log(SPECTER_DEBUG, "Address too long: %s.\n", str);
						return -1;
					}
					strncpy(buff[i], str, EXEC_ARG_SPACE);
					}
					break;
				case SPECTER_IRET_STRING:
					strncpy(buff[i], (char *)iret->value.ptr, EXEC_ARG_SPACE);
					break;
				default:
					specter_log(SPECTER_DEBUG, "Unknown key type %i for \"%s\".\n",
							iret->type, iret->name);
					return -1;
					break;
			}

			/* replace tag with a value */
			argv[ctr] = buff[i++];
		}
		else if (GET_CE(ce,2)->u.value) {
			argv[ctr] = "invalid";
		}
		else {
			specter_log(SPECTER_DEBUG, "Required key \"%s\" invalid.\n",
					iret->name);
			return 0;
		}
	}

	specter_log(SPECTER_DEBUG, "Executing command \"%s\".\n", argv[0]);

	/* execute */
	i = fork();
	if (i == -1) {
		specter_log(SPECTER_ERROR, "Couldn't fork(): %s.\n",
				strerror(errno));
		return -1;
	}
	else if (i) { /* parent */
		if (GET_CE(ce,1)->u.value) {
			if (waitpid(i, &status, 0x0) != -1)
				check_status(argv[0], i, status);
		}
	}
	else { /* child */
		static char *envp[] = { NULL };
		if (execve(argv[0], argv, envp) == -1) {
			specter_log(SPECTER_ERROR, "Couldn't execve(): %s.\n",
					strerror(errno));
		}
	}

	return 0;
}


static specter_output_t exec_op = {
	.name = "exec",
	.ce_base = my_config,
	.ce_num = 3,
	.init = &exec_init,
	.fini = &exec_fini,
	.output = &output_exec,
};


void _init(void)
{
	if (register_output(&exec_op, 0) == -1) {
		specter_log(SPECTER_FATAL, "Can't register.\n");
		exit(EXIT_FAILURE);
	}
}

