/* specter_MYSQL.c
 *
 * specter output plugin for logging to a MySQL database
 *
 * (C) 2000-2001 by Harald Welte <laforge@gnumonks.org>
 * 
 *  15 May 2001, Alex Janssen <alex@ynfonatic.de>:
 *      Added a compability option for older MySQL-servers, which
 *      don't support mysql_real_escape_string
 *
 *  17 May 2001, Alex Janssen <alex@ynfonatic.de>:
 *      Added the --with-mysql-log-ip-as-string feature. This will log
 *      IP's as string rather than an unsigned long integer to the database.
 *	See specter/doc/mysql.table.ipaddr-as-string as an example.
 *	BE WARNED: This has _WAY_ less performance during table searches.
 *
 *  14 Jun 2004, Michal Kwiatkowski <ruby@joker.linuxstuff.pl>:
 *      Fixed for specter.
 *      If iret not found, simply exit.
 */

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 
 *  as published by the Free Software Foundation
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <specter/specter.h>
#include <conffile/conffile.h>
#include <mysql/mysql.h>
#include "lret.h"

#ifdef IP_AS_STRING
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#endif

/* our configuration directives */
static config_entry_t my_config[] = {
	{ .key = "mysqldb", .type = CONFIG_TYPE_STRING, .options = CONFIG_OPT_MANDATORY },
	{ .key = "mysqlhost", .type = CONFIG_TYPE_STRING, .options = CONFIG_OPT_MANDATORY },
	{ .key = "mysqluser", .type = CONFIG_TYPE_STRING, .options = CONFIG_OPT_MANDATORY },
	{ .key = "mysqlpass", .type = CONFIG_TYPE_STRING, .options = CONFIG_OPT_MANDATORY },
	{ .key = "mysqltable", .type = CONFIG_TYPE_STRING, .options = CONFIG_OPT_MANDATORY },
};

struct _field {
	struct _field *next;
	char name[SPECTER_IRET_NAME_LEN];
	specter_iret_t *iret;
};

struct my_data {
	/* the database handle we are using */
	MYSQL *dbh;
	/* a linked list of the fields the table has */
	struct _field *fields;
	/* buffer for our insert statement */
	char *stmt;
	/* pointer to the beginning of the "VALUES" part */
	char *stmt_val;
	/* pointer to current inser position in statement */
	char *stmt_ins;
};


/* our main output function, called by specter */
static int output_mysql(config_entry_t *ce, void *data)
{
	struct my_data *md = data;
	struct _field *f;

	md->stmt_ins = md->stmt_val;

	for (f = md->fields; f; f = f->next) {
		if (!f->iret || !IS_VALID(f->iret)) {
			/* no result, we have to fake something */
			sprintf(md->stmt_ins, "NULL,");
			md->stmt_ins = md->stmt + strlen(md->stmt);
			continue;
		}

		switch (f->iret->type) {
			case SPECTER_IRET_INT8:
				sprintf(md->stmt_ins, "%d,", f->iret->value.i8);
				break;
			case SPECTER_IRET_INT16:
				sprintf(md->stmt_ins, "%d,", f->iret->value.i16);
				break;
			case SPECTER_IRET_INT32:
				sprintf(md->stmt_ins, "%d,", f->iret->value.i32);
				break;
			case SPECTER_IRET_INT64:
				sprintf(md->stmt_ins, "%lld,", f->iret->value.i64);
				break;
			case SPECTER_IRET_UINT8:
				sprintf(md->stmt_ins, "%u,", f->iret->value.ui8);
				break;
			case SPECTER_IRET_UINT16:
				sprintf(md->stmt_ins, "%u,", f->iret->value.ui16);
				break;
			case SPECTER_IRET_IPADDR:
#ifdef IP_AS_STRING
				{
				char *tmpstr;
				*(md->stmt_ins)++ = '\'';
				tmpstr = inet_ntoa((struct in_addr)
						{ntohl(f->iret->value.ui32)});
#ifdef MYSQL_OLD
				mysql_escape_string(md->stmt_ins, tmpstr,
						    strlen(tmpstr));
#else
				mysql_real_escape_string(md->dbh, md->stmt_ins,
							 tmpstr,
							 strlen(tmpstr));
#endif /* MYSQL_OLD */
                                md->stmt_ins = md->stmt + strlen(md->stmt);
                                sprintf(md->stmt_ins, "',");
				}
                                break;
#endif /* IP_AS_STRING */
				/* EVIL: fallthrough when logging IP as
				 * u_int32_t */
			case SPECTER_IRET_UINT32:
				sprintf(md->stmt_ins, "%u,", f->iret->value.ui32);
				break;
			case SPECTER_IRET_UINT64:
				sprintf(md->stmt_ins, "%llu,", f->iret->value.ui64);
				break;
			case SPECTER_IRET_BOOL:
				sprintf(md->stmt_ins, "'%d',", f->iret->value.b);
				break;
			case SPECTER_IRET_STRING:
				*(md->stmt_ins)++ = '\'';
#ifdef MYSQL_OLD
				mysql_escape_string(md->stmt_ins, f->iret->value.ptr,
					strlen(f->iret->value.ptr));
#else
				mysql_real_escape_string(md->dbh, md->stmt_ins,
					f->iret->value.ptr, strlen(f->iret->value.ptr));
#endif
				md->stmt_ins = md->stmt + strlen(md->stmt);
				sprintf(md->stmt_ins, "',");
			/* sprintf(md->stmt_ins, "'%s',", f->iret->value.ptr); */
				break;
			default:
				specter_log(SPECTER_NOTICE,
					"Unknown type %d for key \"%s\".\n",
					f->iret->type, f->iret->name);
				break;
		}
		md->stmt_ins = md->stmt + strlen(md->stmt);
	}
	*(md->stmt_ins - 1) = ')';
	specter_log(SPECTER_DEBUG, "stmt=#%s#\n", md->stmt);

	/* now we have created our statement, insert it */

	if(mysql_real_query(md->dbh, md->stmt, strlen(md->stmt))) {
		specter_log(SPECTER_ERROR, "Error during insert: %s\n",
				mysql_error(md->dbh));
		return 1;
	}

	return 0;
}

#define MYSQL_INSERTTEMPL   "insert into X (Y) values (Z)"
#define MYSQL_VALSIZE	100

/* create the static part of our insert statement */
static int mysql_createstmt(struct my_data *md, const char *table)
{
	struct _field *f;
	unsigned int size;
	char buf[SPECTER_IRET_NAME_LEN];
	char *underscore;

	if (md->stmt) {
		specter_log(SPECTER_NOTICE, "createstmt() called, but stmt"
			" already existing.\n");	
		return 1;
	}

	/* caclulate the size for the insert statement */
	size = strlen(MYSQL_INSERTTEMPL) + strlen(table);

	for (f = md->fields; f; f = f->next) {
		/* we need space for the key and a comma, as well as
		 * enough space for the values */
		size += strlen(f->name) + 1 + MYSQL_VALSIZE;
	}	

	specter_log(SPECTER_DEBUG, "Allocating %u bytes for statement.\n", size);

	md->stmt = (char *) malloc(size);

	if (!md->stmt) {
		specter_log(SPECTER_ERROR, "Couldn't allocate stmt: %s.\n",
				strerror(errno));
		return 1;
	}

	sprintf(md->stmt, "insert into %s (", table);
	md->stmt_val = md->stmt + strlen(md->stmt);

	for (f = md->fields; f; f = f->next) {
		strncpy(buf, f->name, SPECTER_IRET_NAME_LEN);	
		while ((underscore = strchr(buf, '.')))
			*underscore = '_';
		sprintf(md->stmt_val, "%s,", buf);
		md->stmt_val = md->stmt + strlen(md->stmt);
	}
	*(md->stmt_val - 1) = ')';

	sprintf(md->stmt_val, " values (");
	md->stmt_val = md->stmt + strlen(md->stmt);

	return 0;
}

/* find out which columns the table has */
static int mysql_get_columns(MYSQL *dbh, struct _field **fields, const char *table)
{
	MYSQL_RES *result;
	MYSQL_FIELD *field;
	char buf[SPECTER_IRET_NAME_LEN];
	char *underscore;
	struct _field *f;
	specter_iret_t *iret;

	if (!dbh) 
		return 1;

	result = mysql_list_fields(dbh, table, NULL);
	if (!result)
		return 1;

	while ((field = mysql_fetch_field(result))) {

		/* replace all underscores with dots */
		strncpy(buf, field->name, SPECTER_IRET_NAME_LEN);
		while ((underscore = strchr(buf, '_')))
			*underscore = '.';

		if (!(iret = find_iret(buf)))
			return 1;

		specter_log(SPECTER_DEBUG, "Field '%s' found.\n", buf);

		/* prepend it to the linked list */
		f = (struct _field *) malloc(sizeof *f);
		if (!f) {
			specter_log(SPECTER_ERROR, "Couldn't allocate fields: %s.\n",
					strerror(errno));
			return 1;
		}
		strncpy(f->name, buf, SPECTER_IRET_NAME_LEN);
		f->iret = iret;
		f->next = *fields;
		*fields = f;	
	}

	mysql_free_result(result);
	return 0;
}

/* make connection and select database */
static int mysql_open_db(MYSQL **dbh, char *server, char *user, char *pass, char *db)
{
	*dbh = mysql_init(NULL);
	if (!*dbh)
		return 1;

	if (!mysql_real_connect(*dbh, server, user, pass, db, 0, NULL, 0))
		return 1;

	return 0;
}


static void *init_mysql(config_entry_t *ce)
{
	struct my_data *data;

	if ((data = malloc(sizeof(struct my_data))) == NULL) {
		specter_log(SPECTER_FATAL, "Couldn't allocate data: %s.\n",
				strerror(errno));
		return NULL;
	}

	memset(data, 0x0, sizeof(struct my_data));

	if (mysql_open_db(&data->dbh, GET_CE(ce,1)->u.string, GET_CE(ce,2)->u.string, 
			   GET_CE(ce,3)->u.string, GET_CE(ce,0)->u.string)) {
		specter_log(SPECTER_ERROR, "Can't establish database connection.\n");
		return NULL;
	}

	/* read the fieldnames to know which values to insert */
	if (mysql_get_columns(data->dbh, &data->fields, GET_CE(ce,4)->u.string)) {
		specter_log(SPECTER_ERROR, "Unable to get mysql columns.\n");
		return NULL;
	}

	mysql_createstmt(data, GET_CE(ce,4)->u.string);

	return data;
}

static void fini_mysql(config_entry_t *ce, void *data)
{
	struct my_data *md = data;
	struct _field *curr, *next;

	mysql_close(md->dbh);

	/* free linked list of _field structures */
	for (curr = md->fields; curr; curr = next) {
		next = curr->next;
		free(curr);
	}
	free(md->stmt);
	free(data);
}


static specter_output_t mysql_op = {
	.name = "mysql",
	.ce_base = my_config,
	.ce_num = 5,
	.init = &init_mysql,
	.fini = &fini_mysql,
	.output = &output_mysql
};


void _init(void) 
{
	if (register_output(&mysql_op, 0) == -1) {
		specter_log(SPECTER_FATAL, "Can't register.\n");
		exit(EXIT_FAILURE);
	}
}

