/* specter_PGSQL.c
 *
 * specter output plugin for logging to a PGSQL database
 *
 * (C) 2000 by Harald Welte <laforge@gnumonks.org>
 * 
 * This plugin is based on the MySQL plugin made by Harald Welte.
 * The support PostgreSQL were made by Jakab Laszlo.
 *
 *  15 Jun 2004, Michal Kwiatkowski <ruby@joker.linuxstuff.pl>:
 *      Fixed for specter.
 */

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 
 *  as published by the Free Software Foundation
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <specter/specter.h>
#include <conffile/conffile.h>
#include <libpq-fe.h>
#include "lret.h"

#ifdef IP_AS_STRING
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#endif


/* our configuration directives */
static config_entry_t my_config[] = {
	{ .key = "pgsqldb", .type = CONFIG_TYPE_STRING, .options = CONFIG_OPT_MANDATORY },
	{ .key = "pgsqlhost", .type = CONFIG_TYPE_STRING, .options = CONFIG_OPT_MANDATORY },
	{ .key = "pgsqluser", .type = CONFIG_TYPE_STRING, .options = CONFIG_OPT_MANDATORY },
	{ .key = "pgsqlpass", .type = CONFIG_TYPE_STRING, .options = CONFIG_OPT_MANDATORY },
	{ .key = "pgsqltable", .type = CONFIG_TYPE_STRING, .options = CONFIG_OPT_MANDATORY },
};

struct _field {
	struct _field *next;
	char name[SPECTER_IRET_NAME_LEN];
	specter_iret_t *iret;
};

struct my_data {
	/* the database handle we are using */
	PGconn *dbh;
	/* a linked list of the fields the table has */
	struct _field *fields;
	/* buffer for our insert statement */
	char *stmt;
	/* pointer to the beginning of the "VALUES" part */
	char *stmt_val;
	/* pointer to current inser position in statement */
	char *stmt_ins;
};


/* our main output function, called by specter */
static int pgsql_output(config_entry_t *ce, void *data)
{
	struct my_data *md = data;
	struct _field *f;
	PGresult *pgres;

	md->stmt_ins = md->stmt_val;

	for (f = md->fields; f; f = f->next) {
		if (!f->iret) {
			specter_log(SPECTER_NOTICE,
				"No result for %s ?!?\n", f->name);
		}
			
		if (!f->iret || !IS_VALID(f->iret)) {
			/* no result, we have to fake something */
			sprintf(md->stmt_ins, "NULL,");
			md->stmt_ins = md->stmt + strlen(md->stmt);
			continue;
		}
		
		switch (f->iret->type) {
			case SPECTER_IRET_INT8:
				sprintf(md->stmt_ins, "%d,", f->iret->value.i8);
				break;
			case SPECTER_IRET_INT16:
				sprintf(md->stmt_ins, "%d,", f->iret->value.i16);
				break;
			case SPECTER_IRET_INT32:
				sprintf(md->stmt_ins, "%d,", f->iret->value.i32);
				break;
			case SPECTER_IRET_INT64:
				sprintf(md->stmt_ins, "%lld,", f->iret->value.i64);
				break;
			case SPECTER_IRET_UINT8:
				sprintf(md->stmt_ins, "%u,", f->iret->value.ui8);
				break;
			case SPECTER_IRET_UINT16:
				sprintf(md->stmt_ins, "%u,", f->iret->value.ui16);
				break;
			case SPECTER_IRET_IPADDR:
#ifdef IP_AS_STRING
				{
				char *tmpstr;
				*(md->stmt_ins)++ = '\'';
				tmpstr = inet_ntoa((struct in_addr)
						{ntohl(f->iret->value.ui32)});
				PQescapeString(md->stmt_ins, tmpstr, strlen(tmpstr));
				md->stmt_ins = md->stmt + strlen(md->stmt);
				sprintf(md->stmt_ins, "',");
				}
				break;
#endif /* IP_AS_STRING */
				/* EVIL: fallthrough when logging IP as
				 * u_int32_t */
			case SPECTER_IRET_UINT32:
				sprintf(md->stmt_ins, "%u,", f->iret->value.ui32);
				break;
			case SPECTER_IRET_UINT64:
				sprintf(md->stmt_ins, "%llu,", f->iret->value.ui64);
				break;
			case SPECTER_IRET_BOOL:
				sprintf(md->stmt_ins, "'%d',", f->iret->value.b);
				break;
			case SPECTER_IRET_STRING:
				*(md->stmt_ins)++ = '\'';
				PQescapeString(md->stmt_ins, f->iret->value.ptr,
						strlen(f->iret->value.ptr));
				md->stmt_ins = md->stmt + strlen(md->stmt);
				sprintf(md->stmt_ins, "',");
				break;
			case SPECTER_IRET_RAW:
				specter_log(SPECTER_NOTICE, "%s: pgsql doesn't support type RAW.\n",
						f->iret->name);
				sprintf(md->stmt_ins, "NULL,");
				break;
			default:
				specter_log(SPECTER_NOTICE,
					"Unknown type %d for \"%s\".\n",
					f->iret->type, f->iret->name);
				break;
		}
		md->stmt_ins = md->stmt + strlen(md->stmt);
	}
	*(md->stmt_ins - 1) = ')';
	specter_log(SPECTER_DEBUG, "stmt=#%s#\n", md->stmt);

	/* now we have created our statement, insert it */
	/* Added code by Jaki */
	pgres = PQexec(md->dbh, md->stmt);
	if(!pgres || PQresultStatus(pgres) != PGRES_COMMAND_OK) {
		specter_log(SPECTER_ERROR, "Error during insert: %s\n",
				PQresultErrorMessage(pgres));
		return 1;
	}

	return 0;
}

#define PGSQL_INSERTTEMPL   "insert into X (Y) values (Z)"
#define PGSQL_VALSIZE	100

/* create the static part of our insert statement */
static int pgsql_createstmt(struct my_data *md, const char *table)
{
	struct _field *f;
	unsigned int size;
	char buf[SPECTER_IRET_NAME_LEN];
	char *underscore;

	if (md->stmt) {
		specter_log(SPECTER_NOTICE, "createstmt() called, but stmt"
			" already existing.\n");	
		return 1;
	}

	/* caclulate the size for the insert statement */
	size = strlen(PGSQL_INSERTTEMPL) + strlen(table);

	for (f = md->fields; f; f = f->next) {
		/* we need space for the key and a comma, as well as
		 * enough space for the values */
		size += strlen(f->name) + 1 + PGSQL_VALSIZE;
	}	

	specter_log(SPECTER_DEBUG, "Allocating %u bytes for statement.\n", size);

	md->stmt = (char *) malloc(size);

	if (!md->stmt) {
		specter_log(SPECTER_ERROR, "Couldn't allocate stmt: %s.\n",
				strerror(errno));
		return 1;
	}

	sprintf(md->stmt, "insert into %s (", table);
	md->stmt_val = md->stmt + strlen(md->stmt);

	for (f = md->fields; f; f = f->next) {
		strncpy(buf, f->name, SPECTER_IRET_NAME_LEN);	
		while ((underscore = strchr(buf, '.')))
			*underscore = '_';
		sprintf(md->stmt_val, "%s,", buf);
		md->stmt_val = md->stmt + strlen(md->stmt);
	}
	*(md->stmt_val - 1) = ')';

	sprintf(md->stmt_val, " values (");
	md->stmt_val = md->stmt + strlen(md->stmt);

	specter_log(SPECTER_DEBUG, "stmt='%s'\n", md->stmt);

	return 0;
}

/* find out which columns the table has */
static int pgsql_get_columns(PGconn *dbh, struct _field **fields, const char *table)
{
	PGresult *result;
	char buf[SPECTER_IRET_NAME_LEN];
	char pgbuf[256];
	char *underscore;
	struct _field *f;
	specter_iret_t *iret;
	int intaux;

	if (!dbh) 
		return 1;

	strcpy(pgbuf,  "SELECT  a.attname FROM pg_class c, pg_attribute a WHERE c.relname ='");
	strncat(pgbuf, table, strlen(table));
	strcat(pgbuf, "' AND a.attnum>0 AND a.attrelid=c.oid ORDER BY a.attnum");
	specter_log(SPECTER_DEBUG, pgbuf);

	result = PQexec(dbh, pgbuf);
	if(!result) {
		specter_log(SPECTER_DEBUG, "\n result false");
		return 1;
	}

	if (PQresultStatus(result) != PGRES_TUPLES_OK) {
		specter_log(SPECTER_DEBUG, "\n pres_command_not_ok");
		return 1;
	}

	for (intaux = 0; intaux < PQntuples(result); intaux++) {
		/* replace all underscores with dots */
		strncpy(buf, PQgetvalue(result, intaux, 0), SPECTER_IRET_NAME_LEN);
		while ((underscore = strchr(buf, '_')))
			*underscore = '.';

		specter_log(SPECTER_DEBUG, "Field '%s' found: ", buf);

		if (!(iret = find_iret(buf))) {
			specter_log(SPECTER_DEBUG, " no keyid!\n");
			continue;
		}

		specter_log(SPECTER_DEBUG, "keyid %u.\n", iret->name);

		/* prepend it to the linked list */
		f = (struct _field *) malloc(sizeof *f);
		if (!f) {
			specter_log(SPECTER_ERROR, "Couldn't allocate fields: %s.\n",
					strerror(errno));
			return 1;
		}
		strncpy(f->name, buf, SPECTER_IRET_NAME_LEN);
		f->iret = iret;
		f->next = *fields;
		*fields = f;	
	}

	PQclear(result);
	return 0;
}


int exit_nicely(PGconn *conn)
{
	PQfinish(conn);
	return 0;;
}


/* make connection and select database */
static int pgsql_open_db(PGconn **dbh, char *server, char *user, char *pass, char *db)
{
	char connstr[80];

	if (server) {
		strcpy(connstr, " host=");
		strcat(connstr, server);
	}

	strcat(connstr, " dbname=");
	strcat(connstr, db);
	strcat(connstr, " user=");
	strcat(connstr, user);

	if (pass) {
		strcat(connstr, " password=");
		strcat(connstr, pass);
	}

	*dbh = PQconnectdb(connstr);
	if (PQstatus(*dbh) != CONNECTION_OK) {
		exit_nicely(*dbh);
		return 1;
	}

	return 0;
}


static void *pgsql_init(config_entry_t *ce)
{
	struct my_data *data;

	if ((data = malloc(sizeof(struct my_data))) == NULL) {
		specter_log(SPECTER_FATAL, "Couldn't allocate data: %s.\n",
				strerror(errno));
		return NULL;
	}

	memset(data, 0x0, sizeof(struct my_data));

	if (pgsql_open_db(&data->dbh, GET_CE(ce,1)->u.string, GET_CE(ce,2)->u.string, 
			   GET_CE(ce,3)->u.string, GET_CE(ce,0)->u.string)) {
		specter_log(SPECTER_ERROR, "Can't establish database connection.\n");
		return NULL;
	}

	/* read the fieldnames to know which values to insert */
	if (pgsql_get_columns(data->dbh, &data->fields, GET_CE(ce,4)->u.string)) {
		specter_log(SPECTER_ERROR, "Unable to get pgsql columns.\n");
		return NULL;
	}

	pgsql_createstmt(data, GET_CE(ce,4)->u.string);

	return data;
}


static void pgsql_fini(config_entry_t *ce, void *data)
{
	struct my_data *md = data;

	PQfinish(md->dbh);
}


static specter_output_t pgsql_op = {
	.name = "pgsql",
	.ce_base = my_config,
	.ce_num = 5,
	.init = &pgsql_init,
	.fini = &pgsql_fini,
	.output = &pgsql_output
};


void _init(void) 
{
	if (register_output(&pgsql_op, 0) == -1) {
		specter_log(SPECTER_FATAL, "Can't register.\n");
		exit(EXIT_FAILURE);
	}
}

